@@ -49,7 +49,7 @@ class HBChannel(Thread):
 
     def __init__(
         self,
-        context: zmq.asyncio.Context = None,
+        context: t.Optional[zmq.asyncio.Context] = None,
         session: t.Optional[Session] = None,
         address: t.Union[t.Tuple[str, int], str] = "",
     ):
@@ -109,6 +109,7 @@ class HBChannel(Thread):
         loop = asyncio.new_event_loop()
         asyncio.set_event_loop(loop)
         loop.run_until_complete(self._async_run())
+        loop.close()
 
     async def _async_run(self) -> None:
         """The thread's main activity.  Call start() instead."""
@@ -210,7 +211,7 @@ class ZMQSocketChannel(object):
         self.socket: t.Optional[zmq.sugar.socket.Socket] = socket
         self.session = session
 
-    async def _recv(self, **kwargs) -> t.Dict[str, t.Any]:
+    async def _recv(self, **kwargs: t.Any) -> t.Dict[str, t.Any]:
         assert self.socket is not None
         msg = await self.socket.recv_multipart(**kwargs)
         ident, smsg = self.session.feed_identities(msg)
